// compiler avec gcc -g -Wall message_secret.c -o message_secret.ex
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
       
typedef void (*sighandler_t)(int);
#define nb_messages 5


const char * tab_messages[nb_messages] = {
	"Il faut affranchir nos compagnons",
	"Il faut franchir les compagnons",
	"Il faut rafraichir les maquignons",
	"Vous ne connaissez rien au code !!",
	"Il faut que vous fassiez blanchir des champignons"
};

const char * tab_processus[nb_messages] = {
	"Arthur",
	"Caradoc",
	"Perceval",
	"Lancelot",
	"Merlin"
};

/**
 * fonction de traitement des messages
 * @param num : le numero du processus
 * @param message : le message recu du précédant
 * @return le nouveau message
 */
const char* traite_message(int num, const char* message) {
	usleep(rand()%1000000);
	// pour le premier processus le message est simplement le premier de la liste
	if (num == 0) {
		fprintf(stderr, "%s : le message est '%s'\n", tab_processus[num], tab_messages[num]);
		return tab_messages[num];
	}
	// pour les autres, il faut vérifier qu'il n'a pas été modifié
	if (strcmp(tab_messages[num-1], message)!=0) {
		fprintf(stderr, "Attention, le message a été modifié, '%s' != '%s'\n", message, tab_messages[num-1]);
	}
	fprintf(stderr, "%s : le message est '%s'\n", tab_processus[num], tab_messages[num]);
	return tab_messages[num];
}

int cont;
void fctsig(int s) {
	fprintf(stderr, "signal recu %d et %d\n", s, SIGTERM);
	if (s == SIGINT) {
			// rien a faire fprintf(stdout, "Je recommence\n");
	} else if (s == SIGQUIT) {
		cont = 0;
	}
}

int read_all(int sock, char *data, int taille) {
	int deja = 0;
	int nb;
	
	while(deja < taille) {
		nb = read(sock, data+deja, taille-deja);
		if (nb == -1) {
			perror("read");
			exit(1);
		}
		if (nb == 0) {
			fprintf(stderr, "attention, le fd est coupé alors qu'on attend encore %d octet\n", taille-deja);
			break;
		}
		deja += nb;
	}
	return deja;
}

int write_all(int sock, const char *data, int taille) {
	int deja = 0;
	int nb;
	
	while(deja < taille) {
		nb = write(sock, data+deja, taille-deja);
		if (nb == -1) {
			perror("write");
			exit(1);
		}
		deja += nb;
	}
	return deja;
}



int main(int argc, char *argv[]) {
	int i;
	int res;
	int p[2];
	int in = -1, out = -1;
	// il y a 5 processus donc 4 fils
	int nb_processus = 5;
	pid_t fils;
	
	for (i = 0; i< nb_processus -1; i++) {
		res = pipe(p);
		if (res == -1) {
			perror("pipe");
			exit(1);
		}
		
		fils = fork();
		if (fils == -1) {
			perror("fork");
			exit(1);
		}
		
		if (fils != 0) {
			// c'est le père

			// fermeture du coté lecture
			res = close(p[0]);
			if (res == -1) {
				perror("close");
				exit(1);
			}
			out = p[1];
			break;
		}
		// seul le fils arrive ici
		in = p[0];
		// fermeture du coté écriture
		res = close(p[1]);
		if (res == -1) {
			perror("close");
			exit(1);
		}
		
	}
	int num = i;
	fprintf(stdout, "Je suis %s\n", tab_processus[num]);
	
	sighandler_t s = signal(SIGINT, fctsig);
	if (s == SIG_ERR) {
		perror("signal");
		exit(1);
	}
	s = signal(SIGQUIT, fctsig);
	if (s == SIG_ERR) {
		perror("signal");
		exit(1);
	}
	//~ for (int i = 0; i<30; i++) {
		
	//~ s = signal(i, fctsig);
	//~ if (s == SIG_ERR) {
		//~ fprintf(stderr, "signal %d", i);
	//~ //	exit(1);
	//~ }
	//~ }
	
	int nbmessage = 0;
	cont = 1;
	while(cont) {
		char *buff;
		if (num != 0) {
			// on lit la taille du message puis le message
			int taille;
			res = read_all(in, (char*)&taille, sizeof(int));
			buff = (char*)malloc(taille+1);
			
			res = read_all(in, buff, taille);
			buff[taille] = '\0';
		} else {
			buff = NULL;
		}
		nbmessage++;
		const char *new_message = traite_message(num, buff);
		
		if (num != nb_processus-1) {
			// on envoie le la taille du message puis le message
			int t = strlen(new_message);
			write_all(out, (char*)&t, sizeof(int));
			write_all(out, new_message, t);
		}
		if (num == 0) {
			// on attend le signal
			pause();
			fprintf(stdout, "%s : Je recommence\n", tab_processus[num]);
		}
	}		
	if (in != -1) {
		res = close(in);
		if (res == -1) {
			perror("close");
			exit(1);
		}
	}
	if (out != -1) {
		res = close(out);	
		if (res == -1) {
			perror("close");
			exit(1);
		}
	}
	
	int status;
	if (num != nb_processus-1) {
		res = waitpid(fils, &status, 0);
		if (res == -1) {
			perror("waitpid");
			exit(1);
		}
		if (WIFEXITED(status)) {
			fprintf(stdout, "Nb de message message recu par le fils %d\n", WEXITSTATUS(status));
		}
	}
	if (num != 0) {
		return nbmessage;
	} else {
		return 0;
	}
}
